@@ -1,5 +1,5 @@
 /* Case-insensitive buffer comparator.
-   Copyright (C) 1996 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1997, 2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,9 +15,9 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Jim Meyering (meyering@na-net.ornl.gov) */
+/* Written by Jim Meyering.  */
 
-#ifdef HAVE_CONFIG_H
+#if HAVE_CONFIG_H
 # include <config.h>
 #endif
 
@@ -25,37 +25,36 @@
 #include <ctype.h>
 
 #if defined (STDC_HEADERS) || (!defined (isascii) && !defined (HAVE_ISASCII))
-#define IN_CTYPE_DOMAIN(c) 1
+# define IN_CTYPE_DOMAIN(Char) 1
 #else
-#define IN_CTYPE_DOMAIN(c) isascii(c)
+# define IN_CTYPE_DOMAIN(Char) isascii(Char)
 #endif
-#define ISUPPER(c) (IN_CTYPE_DOMAIN (c) && isupper (c))
+#define ISLOWER(Char) (IN_CTYPE_DOMAIN (Char) && islower (Char))
 
 #if _LIBC || STDC_HEADERS
-# define TOLOWER(c) tolower (c)
+# define TOUPPER(Char) toupper (Char)
 #else
-# define TOLOWER(c) (ISUPPER (c) ? tolower (c) : (c))
+# define TOUPPER(Char) (ISLOWER (Char) ? toupper (Char) : (Char))
 #endif
 
 #include "memcasecmp.h"
 
-/* Like memcmp, but ignore differences in case.  */
+/* Like memcmp, but ignore differences in case.
+   Convert to upper case (not lower) before comparing so that
+   join -i works with sort -f.  */
 
 int
-memcasecmp (vs1, vs2, n)
-     const void *vs1;
-     const void *vs2;
-     size_t n;
+memcasecmp (const void *vs1, const void *vs2, size_t n)
 {
   unsigned int i;
-  unsigned char *s1 = (unsigned char *) vs1;
-  unsigned char *s2 = (unsigned char *) vs2;
+  unsigned char const *s1 = (unsigned char const *) vs1;
+  unsigned char const *s2 = (unsigned char const *) vs2;
   for (i = 0; i < n; i++)
     {
       unsigned char u1 = *s1++;
       unsigned char u2 = *s2++;
-      if (TOLOWER (u1) != TOLOWER (u2))
-        return TOLOWER (u1) - TOLOWER (u2);
+      if (TOUPPER (u1) != TOUPPER (u2))
+        return TOUPPER (u1) - TOUPPER (u2);
     }
   return 0;
 }
