@@ -1,5 +1,5 @@
 /* cat -- concatenate files and print on the standard output.
-   Copyright (C) 88, 90, 91, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 # include <sys/ioctl.h>
 #endif
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 #include "safe-read.h"
 
@@ -54,22 +55,26 @@ static char *infile;
 /* Descriptor on which input file is open.  */
 static int input_desc;
 
-/* Descriptor on which output file is open.  Always is 1.  */
-static int output_desc;
-
-/* Buffer for line numbers.  */
-static char line_buf[13] =
-{' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', '0', '\t', '\0'};
+/* Buffer for line numbers.
+   An 11 digit counter may overflow within an hour on a P2/466,
+   an 18 digit counter needs about 1000y */
+#define LINE_COUNTER_BUF_LEN 20
+static char line_buf[LINE_COUNTER_BUF_LEN] =
+  {
+    ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
+    ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', '0',
+    '\t', '\0'
+  };
 
 /* Position in `line_buf' where printing starts.  This will not change
    unless the number of lines is larger than 999999.  */
-static char *line_num_print = line_buf + 5;
+static char *line_num_print = line_buf + LINE_COUNTER_BUF_LEN - 8;
 
 /* Position of the first digit in `line_buf'.  */
-static char *line_num_start = line_buf + 10;
+static char *line_num_start = line_buf + LINE_COUNTER_BUF_LEN - 3;
 
 /* Position of the last digit in `line_buf'.  */
-static char *line_num_end = line_buf + 10;
+static char *line_num_end = line_buf + LINE_COUNTER_BUF_LEN - 3;
 
 /* Preserves the `cat' function's local `newlines' between invocations.  */
 static int newlines2 = 0;
@@ -131,13 +136,15 @@ next_line_num (void)
       *endp-- = '0';
     }
   while (endp >= line_num_start);
-  *--line_num_start = '1';
+  if (line_num_start > line_buf)
+    *--line_num_start = '1';
+  else
+    *line_buf = '>';
   if (line_num_start < line_num_print)
     line_num_print--;
 }
 
-/* Plain cat.  Copies the file behind `input_desc' to the file behind
-   `output_desc'.  */
+/* Plain cat.  Copies the file behind `input_desc' to STDOUT_FILENO.  */
 
 static void
 simple_cat (
@@ -172,7 +179,7 @@ simple_cat (
 
       /* Write this block out.  */
 
-      if (full_write (output_desc, buf, n_read) < 0)
+      if (full_write (STDOUT_FILENO, buf, n_read) < 0)
 	error (EXIT_FAILURE, errno, _("write error"));
     }
 }
@@ -253,7 +260,7 @@ cat (
 	      unsigned char *wp = outbuf;
 	      do
 		{
-		  if (full_write (output_desc, wp, outsize) < 0)
+		  if (full_write (STDOUT_FILENO, wp, outsize) < 0)
 		    error (EXIT_FAILURE, errno, _("write error"));
 		  wp += outsize;
 		}
@@ -306,7 +313,7 @@ cat (
 		{
 		  int n_write = bpout - outbuf;
 
-		  if (full_write (output_desc, outbuf, n_write) < 0)
+		  if (full_write (STDOUT_FILENO, outbuf, n_write) < 0)
 		    error (EXIT_FAILURE, errno, _("write error"));
 		  bpout = outbuf;
 		}
@@ -343,13 +350,21 @@ cat (
 
 	      if (++newlines > 0)
 		{
-		  /* Are multiple adjacent empty lines to be substituted by
-		     single ditto (-s), and this was the second empty line?  */
-
-		  if (squeeze_empty_lines && newlines >= 2)
+		  if (newlines >= 2)
 		    {
-		      ch = *bpin++;
-		      continue;
+		      /* Limit this to 2 here.  Otherwise, with lots of
+			 consecutive newlines, the counter could wrap
+			 around at INT_MAX.  */
+		      newlines = 2;
+
+		      /* Are multiple adjacent empty lines to be substituted
+			 by single ditto (-s), and this was the second empty
+			 line?  */
+		      if (squeeze_empty_lines)
+			{
+			  ch = *bpin++;
+			  continue;
+			}
 		    }
 
 		  /* Are line numbers to be written at empty lines (-n)?  */
@@ -537,6 +552,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   /* Parse command line options.  */
 
   while ((c = getopt_long (argc, argv,
@@ -622,11 +639,9 @@ main (int argc, char **argv)
 	}
     }
 
-  output_desc = 1;
-
   /* Get device, i-node number, and optimal blocksize of output.  */
 
-  if (fstat (output_desc, &stat_buf) < 0)
+  if (fstat (STDOUT_FILENO, &stat_buf) < 0)
     error (EXIT_FAILURE, errno, _("standard output"));
 
   outsize = ST_BLKSIZE (stat_buf);
@@ -659,13 +674,13 @@ main (int argc, char **argv)
      with only CR-LF is an empty line.  (Besides, if they ask for
      one of these options, they don't care much about the original
      file contents anyway).  */
-  if ((!isatty (output_desc)
+  if ((!isatty (STDOUT_FILENO)
        && !(numbers || squeeze_empty_lines || mark_line_ends))
       || binary_files)
     {
       /* Switch stdout to BINARY mode.  */
       binary_output = 1;
-      SET_BINARY (output_desc);
+      SET_BINARY (STDOUT_FILENO);
       /* When stdout is in binary mode, make sure all input files are
 	 also read in binary mode.  */
       file_open_mode |= O_BINARY;
@@ -681,8 +696,8 @@ main (int argc, char **argv)
 
       /* Setting stdin to binary switches the console device to
 	 raw I/O, which also affects stdout to console.  Undo that.  */
-      if (isatty (output_desc))
-	setmode (output_desc, O_TEXT);
+      if (isatty (STDOUT_FILENO))
+	setmode (STDOUT_FILENO, O_TEXT);
     }
 #endif
 
@@ -715,7 +730,7 @@ main (int argc, char **argv)
 		 This is so "cat > xyzzy" creates a DOS-style text
 		 file, like people expect.  */
 	      if (tty_in && optind <= argc)
-		setmode (output_desc, O_TEXT);
+		setmode (STDOUT_FILENO, O_TEXT);
 	      else
 		{
 		  SET_BINARY (input_desc);
@@ -756,7 +771,7 @@ main (int argc, char **argv)
 
       if (check_redirection
 	  && stat_buf.st_dev == out_dev && stat_buf.st_ino == out_ino
-	  && (input_desc != STDIN_FILENO || output_desc != STDOUT_FILENO))
+	  && (input_desc != STDIN_FILENO))
 	{
 	  error (0, 0, _("%s: input file is output file"), infile);
 	  exit_status = 1;
@@ -778,8 +793,8 @@ main (int argc, char **argv)
 	{
 	  inbuf = (unsigned char *) xmalloc (insize + 1);
 
-	  /* Why are (OUTSIZE  - 1 + INSIZE * 4 + 13) bytes allocated for
-	     the output buffer?
+	  /* Why are (OUTSIZE - 1 + INSIZE * 4 + LINE_COUNTER_BUF_LEN)
+	     bytes allocated for the output buffer?
 
 	     A test whether output needs to be written is done when the input
 	     buffer empties or when a newline appears in the input.  After
@@ -787,14 +802,16 @@ main (int argc, char **argv)
 	     buffer.  Now INSIZE bytes of input is read.  Each input character
 	     may grow by a factor of 4 (by the prepending of M-^).  If all
 	     characters do, and no newlines appear in this block of input, we
-	     will have at most (OUTSIZE - 1 + INSIZE) bytes in the buffer.  If
-	     the last character in the preceding block of input was a
+	     will have at most (OUTSIZE - 1 + INSIZE * 4) bytes in the buffer.
+	     If the last character in the preceding block of input was a
 	     newline, a line number may be written (according to the given
 	     options) as the first thing in the output buffer. (Done after the
-	     new input is read, but before processing of the input begins.)  A
-	     line number requires seldom more than 13 positions.  */
+	     new input is read, but before processing of the input begins.)
+	     A line number requires seldom more than LINE_COUNTER_BUF_LEN
+	     positions.  */
 
-	  outbuf = (unsigned char *) xmalloc (outsize - 1 + insize * 4 + 13);
+	  outbuf = (unsigned char *) xmalloc (outsize - 1 + insize * 4
+					      + LINE_COUNTER_BUF_LEN);
 
 	  cat (inbuf, insize, outbuf, outsize, quote,
 	       output_tabs, numbers, numbers_at_empty_lines, mark_line_ends,
@@ -816,8 +833,6 @@ main (int argc, char **argv)
 
   if (have_read_stdin && close (0) < 0)
     error (EXIT_FAILURE, errno, "-");
-  if (close (1) < 0)
-    error (EXIT_FAILURE, errno, _("write error"));
 
   exit (exit_status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
