@@ -1,5 +1,5 @@
 /* cksum -- calculate and print POSIX.2 checksums and sizes of files
-   Copyright (C) 92, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 92, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -113,6 +113,7 @@ main ()
 # include <getopt.h>
 # include <sys/types.h>
 # include "system.h"
+# include "closeout.h"
 # include "long-options.h"
 # include "error.h"
 
@@ -192,7 +193,7 @@ static int have_read_stdin;
    Return 0 if successful, -1 if an error occurs. */
 
 static int
-cksum (char *file, int print_name)
+cksum (const char *file, int print_name)
 {
   unsigned char buf[BUFLEN];
   unsigned long crc = 0;
@@ -267,9 +268,10 @@ usage (int status)
   else
     {
       printf (_("\
-Usage: %s [OPTION]... [FILE]...\n\
+Usage: %s [FILE]...\n\
+  or:  %s [OPTION]\n\
 "),
-	      program_name);
+	      program_name, program_name);
       printf (_("\
 Print CRC checksum and byte counts of each FILE.\n\
 \n\
@@ -292,6 +294,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   parse_long_options (argc, argv, PROGRAM_NAME, GNU_PACKAGE, VERSION,
 		      AUTHORS, usage);
 
