@@ -1,63 +1,26 @@
 /* cut - remove parts of lines of files
-   Copyright (C) 1984, 1997, 1998, 1999 by David M. Ihnat
+   Copyright (C) 1984, 1997, 1998, 1999, 2000 by David M. Ihnat
 
-   This program is a total rewrite of the Bell Laboratories Unix(Tm)
-   command of the same name, as of System V.  It contains no proprietary
-   code, and therefore may be used without violation of any proprietary
-   agreements whatsoever.  However, you will notice that the program is
-   copyrighted by me.  This is to assure the program does *not* fall
-   into the public domain.  Thus, I may specify just what I am now:
-   This program may be freely copied and distributed, provided this notice
-   remains; it may not be sold for profit without express written consent of
-   the author.
-   Please note that I recreated the behavior of the Unix(Tm) 'cut' command
-   as faithfully as possible; however, I haven't run a full set of regression
-   tests.  Thus, the user of this program accepts full responsibility for any
-   effects or loss; in particular, the author is not responsible for any losses,
-   explicit or incidental, that may be incurred through use of this program.
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
 
-   I ask that any bugs (and, if possible, fixes) be reported to me when
-   possible.  -David Ihnat (312) 784-4544 ignatz@homebru.chi.il.us
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-   POSIX changes, bug fixes, long-named options, and cleanup
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+
+/* Written by David Ihnat.  */
+
+/* POSIX changes, bug fixes, long-named options, and cleanup
    by David MacKenzie <djm@gnu.ai.mit.edu>.
 
-   Rewrite cut_fields and cut_bytes -- Jim Meyering (meyering@comco.com).
-
-   Options:
-   --bytes=byte-list
-   -b byte-list			Print only the bytes in positions listed
-				in BYTE-LIST.
-				Tabs and backspaces are treated like any
-				other character; they take up 1 byte.
-
-   --characters=character-list
-   -c character-list		Print only characters in positions listed
-				in CHARACTER-LIST.
-				The same as -b for now, but
-				internationalization will change that.
-				Tabs and backspaces are treated like any
-				other character; they take up 1 character.
-
-   --fields=field-list
-   -f field-list		Print only the fields listed in FIELD-LIST.
-				Fields are separated by a TAB by default.
-
-   --delimiter=delim
-   -d delim			For -f, fields are separated by the first
-				character in DELIM instead of TAB.
-
-   -n				Do not split multibyte chars (no-op for now).
-
-   --only-delimited
-   -s				For -f, do not print lines that do not contain
-				the field separator character.
-
-   The BYTE-LIST, CHARACTER-LIST, and FIELD-LIST are one or more numbers
-   or ranges separated by commas.  The first byte, character, and field
-   are numbered 1.
-
-   A FILE of `-' means standard input. */
+   Rewrite cut_fields and cut_bytes -- Jim Meyering.  */
 
 #include <config.h>
 
@@ -66,6 +29,8 @@
 #include <getopt.h>
 #include <sys/types.h>
 #include "system.h"
+#include "getstr.h"
+#include "closeout.h"
 #include "error.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
@@ -73,8 +38,6 @@
 
 #define AUTHORS "David Ihnat, David MacKenzie, and Jim Meyering"
 
-char *xstrdup ();
-
 #define FATAL_ERROR(Message)						\
   do									\
     {									\
@@ -118,7 +81,7 @@ struct range_pair
 static char *field_1_buffer;
 
 /* The number of bytes allocated for FIELD_1_BUFFER.  */
-static int field_1_bufsize;
+static size_t field_1_bufsize;
 
 /* The largest field or byte index used as an endpoint of a closed
    or degenerate range specification;  this doesn't include the starting
@@ -173,6 +136,13 @@ static char *output_delimiter_string;
 /* Nonzero if we have ever read standard input. */
 static int have_read_stdin;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  OUTPUT_DELIMITER_OPTION = CHAR_MAX + 1
+};
+
 static struct option const longopts[] =
 {
   {"bytes", required_argument, 0, 'b'},
@@ -180,7 +150,7 @@ static struct option const longopts[] =
   {"fields", required_argument, 0, 'f'},
   {"delimiter", required_argument, 0, 'd'},
   {"only-delimited", no_argument, 0, 's'},
-  {"output-delimiter", required_argument, 0, CHAR_MAX + 1},
+  {"output-delimiter", required_argument, 0, OUTPUT_DELIMITER_OPTION},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {0, 0, 0, 0}
@@ -204,7 +174,9 @@ Print selected parts of lines from each FILE to standard output.\n\
   -b, --bytes=LIST        output only these bytes\n\
   -c, --characters=LIST   output only these characters\n\
   -d, --delimiter=DELIM   use DELIM instead of TAB for field delimiter\n\
-  -f, --fields=LIST       output only these fields\n\
+  -f, --fields=LIST       output only these fields;  also print any line\n\
+                            that contains no delimiter character, unless\n\
+                            the -s option is specified\n\
   -n                      (ignored)\n\
   -s, --only-delimited    do not print lines not containing delimiters\n\
       --output-delimiter=STRING  use STRING as the output delimiter\n\
@@ -227,89 +199,6 @@ With no FILE, or when FILE is -, read standard input.\n\
   exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
 
-/* The following function was copied from getline.c, but with these changes:
-   - Read up to and including a newline or TERMINATOR, whichever comes first.
-   The original does not treat newline specially.
-   - Remove unused argument, OFFSET.
-   - Use xmalloc and xrealloc instead of malloc and realloc.
-   - Declare this function static.  */
-
-/* Always add at least this many bytes when extending the buffer.  */
-#define MIN_CHUNK 64
-
-/* Read up to (and including) a newline or TERMINATOR from STREAM into
-   *LINEPTR (and null-terminate it). *LINEPTR is a pointer returned from
-   xmalloc (or NULL), pointing to *N characters of space.  It is
-   xrealloc'd as necessary.  Return the number of characters read (not
-   including the null terminator), or -1 on error or EOF.  */
-
-static int
-getstr (char **lineptr, int *n, FILE *stream, int terminator)
-{
-  int nchars_avail;		/* Allocated but unused chars in *LINEPTR.  */
-  char *read_pos;		/* Where we're reading into *LINEPTR. */
-
-  if (!lineptr || !n || !stream)
-    return -1;
-
-  if (!*lineptr)
-    {
-      *n = MIN_CHUNK;
-      *lineptr = (char *) xmalloc (*n);
-      if (!*lineptr)
-	return -1;
-    }
-
-  nchars_avail = *n;
-  read_pos = *lineptr;
-
-  for (;;)
-    {
-      register int c = getc (stream);
-
-      /* We always want at least one char left in the buffer, since we
-	 always (unless we get an error while reading the first char)
-	 NUL-terminate the line buffer.  */
-
-      assert (*n - nchars_avail == read_pos - *lineptr);
-      if (nchars_avail < 1)
-	{
-	  if (*n > MIN_CHUNK)
-	    *n *= 2;
-	  else
-	    *n += MIN_CHUNK;
-
-	  nchars_avail = *n + *lineptr - read_pos;
-	  *lineptr = xrealloc (*lineptr, *n);
-	  if (!*lineptr)
-	    return -1;
-	  read_pos = *n - nchars_avail + *lineptr;
-	  assert (*n - nchars_avail == read_pos - *lineptr);
-	}
-
-      if (feof (stream) || ferror (stream))
-	{
-	  /* Return partial line, if any.  */
-	  if (read_pos == *lineptr)
-	    return -1;
-	  else
-	    break;
-	}
-
-      *read_pos++ = c;
-      nchars_avail--;
-
-      if (c == terminator || c == '\n')
-	/* Return the line.  */
-	break;
-    }
-
-  /* Done - NUL terminate and return the number of chars read.  */
-  *read_pos = '\0';
-
-  return read_pos - *lineptr;
-}
-
 static int
 print_kth (unsigned int k)
 {
@@ -552,9 +441,14 @@ cut_fields (FILE *stream)
 	{
 	  int len;
 
-	  len = getstr (&field_1_buffer, &field_1_bufsize, stream, delim);
+	  len = getstr (&field_1_buffer, &field_1_bufsize, stream,
+			delim, '\n', 0);
 	  if (len < 0)
-	    break;
+	    {
+	      if (ferror (stream) || feof (stream))
+		break;
+	      xalloc_die ();
+	    }
 
 	  assert (len != 0);
 
@@ -695,6 +589,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   operating_mode = undefined_mode;
 
   /* By default, all non-delimited lines are printed.  */
@@ -738,7 +634,7 @@ main (int argc, char **argv)
 	  delim_specified = 1;
 	  break;
 
-	case CHAR_MAX + 1:
+	case OUTPUT_DELIMITER_OPTION:
 	  /* Interpret --output-delimiter='' to mean
 	     `use the NUL byte as the delimiter.'  */
 	  output_delimiter_length = (optarg[0] == '\0'
@@ -795,8 +691,6 @@ main (int argc, char **argv)
       error (0, errno, "-");
       exit_status = 1;
     }
-  if (ferror (stdout) || fclose (stdout) == EOF)
-    error (EXIT_FAILURE, errno, _("write error"));
 
   exit (exit_status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
