@@ -65,11 +65,6 @@ typedef double LONG_DOUBLE;
 # define LDBL_DIG DBL_DIG
 #endif
 
-#if !HAVE_FSEEKO
-# undef fseeko
-# define fseeko(Stream, Offset, Whence) (-1)
-#endif
-
 enum size_spec
   {
     NO_SIZE,
@@ -158,16 +153,17 @@ static const char *const charname[33] =
 /* A printf control string for printing a file offset.  */
 static const char *output_address_fmt_string;
 
-/* FIXME: make this the number of octal digits in an unsigned long.  */
-#define MAX_ADDRESS_LENGTH 13
+/* The number of octal digits required to represent the largest off_t value.  */
+#define MAX_ADDRESS_LENGTH \
+  ((sizeof (off_t) * CHAR_BIT + CHAR_BIT - 1) / 3)
 
 /* Space for a normal address, a space, a pseudo address, parentheses
    around the pseudo address, and a trailing zero byte. */
 static char address_fmt_buffer[2 * MAX_ADDRESS_LENGTH + 4];
 static char address_pad[MAX_ADDRESS_LENGTH + 1];
 
-static unsigned long int string_min;
-static unsigned long int flag_dump_strings;
+static size_t string_min;
+static int flag_dump_strings;
 
 /* Non-zero if we should recognize the pre-POSIX non-option arguments
    that specified at most one file and optional arguments specifying
@@ -183,7 +179,7 @@ static long int pseudo_offset;
 
 /* Function to format an address and optionally an additional parenthesized
    pseudo-address; it returns the formatted string.  */
-static const char *(*format_address) PARAMS ((long unsigned int));
+static const char *(*format_address) PARAMS ((off_t));
 
 /* The number of input bytes to skip before formatting and writing.  */
 static off_t n_bytes_to_skip = 0;
@@ -364,10 +360,9 @@ lcm (unsigned int u, unsigned int v)
 }
 
 static void
-print_s_char (long unsigned int n_bytes, const char *block,
-	      const char *fmt_string)
+print_s_char (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes; i > 0; i--)
     {
       int tmp = (unsigned) *(const unsigned char *) block;
@@ -380,10 +375,9 @@ print_s_char (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_char (long unsigned int n_bytes, const char *block,
-	    const char *fmt_string)
+print_char (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes; i > 0; i--)
     {
       unsigned int tmp = *(const unsigned char *) block;
@@ -393,10 +387,9 @@ print_char (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_s_short (long unsigned int n_bytes, const char *block,
-	       const char *fmt_string)
+print_s_short (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (unsigned short); i > 0; i--)
     {
       int tmp = (unsigned) *(const unsigned short *) block;
@@ -409,10 +402,9 @@ print_s_short (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_short (long unsigned int n_bytes, const char *block,
-	     const char *fmt_string)
+print_short (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (unsigned short); i > 0; i--)
     {
       unsigned int tmp = *(const unsigned short *) block;
@@ -422,10 +414,9 @@ print_short (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_int (long unsigned int n_bytes, const char *block,
-	   const char *fmt_string)
+print_int (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (unsigned int); i > 0; i--)
     {
       unsigned int tmp = *(const unsigned int *) block;
@@ -435,10 +426,9 @@ print_int (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_long (long unsigned int n_bytes, const char *block,
-	    const char *fmt_string)
+print_long (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (unsigned long); i > 0; i--)
     {
       unsigned long tmp = *(const unsigned long *) block;
@@ -449,10 +439,9 @@ print_long (long unsigned int n_bytes, const char *block,
 
 #ifdef HAVE_UNSIGNED_LONG_LONG
 static void
-print_long_long (long unsigned int n_bytes, const char *block,
-		 const char *fmt_string)
+print_long_long (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (unsigned long long); i > 0; i--)
     {
       unsigned long long tmp = *(const unsigned long long *) block;
@@ -463,10 +452,9 @@ print_long_long (long unsigned int n_bytes, const char *block,
 #endif
 
 static void
-print_float (long unsigned int n_bytes, const char *block,
-	     const char *fmt_string)
+print_float (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (float); i > 0; i--)
     {
       float tmp = *(const float *) block;
@@ -476,10 +464,9 @@ print_float (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_double (long unsigned int n_bytes, const char *block,
-	      const char *fmt_string)
+print_double (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (double); i > 0; i--)
     {
       double tmp = *(const double *) block;
@@ -490,10 +477,9 @@ print_double (long unsigned int n_bytes, const char *block,
 
 #ifdef HAVE_LONG_DOUBLE
 static void
-print_long_double (long unsigned int n_bytes, const char *block,
-		   const char *fmt_string)
+print_long_double (off_t n_bytes, const char *block, const char *fmt_string)
 {
-  int i;
+  off_t i;
   for (i = n_bytes / sizeof (LONG_DOUBLE); i > 0; i--)
     {
       LONG_DOUBLE tmp = *(const LONG_DOUBLE *) block;
@@ -505,9 +491,9 @@ print_long_double (long unsigned int n_bytes, const char *block,
 #endif
 
 static void
-dump_hexl_mode_trailer (long unsigned int n_bytes, const char *block)
+dump_hexl_mode_trailer (off_t n_bytes, const char *block)
 {
-  int i;
+  off_t i;
   fputs ("  >", stdout);
   for (i = n_bytes; i > 0; i--)
     {
@@ -520,10 +506,10 @@ dump_hexl_mode_trailer (long unsigned int n_bytes, const char *block)
 }
 
 static void
-print_named_ascii (long unsigned int n_bytes, const char *block,
-		   const char *unused_fmt_string)
+print_named_ascii (off_t n_bytes, const char *block,
+		   const char *unused_fmt_string ATTRIBUTE_UNUSED)
 {
-  int i;
+  off_t i;
   for (i = n_bytes; i > 0; i--)
     {
       unsigned int c = *(const unsigned char *) block;
@@ -547,10 +533,10 @@ print_named_ascii (long unsigned int n_bytes, const char *block,
 }
 
 static void
-print_ascii (long unsigned int n_bytes, const char *block,
-	     const char *unused_fmt_string)
+print_ascii (off_t n_bytes, const char *block,
+	     const char *unused_fmt_string ATTRIBUTE_UNUSED)
 {
-  int i;
+  off_t i;
   for (i = n_bytes; i > 0; i--)
     {
       unsigned int c = *(const unsigned char *) block;
@@ -652,7 +638,7 @@ decode_one_format (const char *s_orig, const char *s, const char **next,
   enum output_format fmt;
   const char *pre_fmt_string;
   char *fmt_string;
-  void (*print_function) ();
+  void (*print_function) PARAMS ((off_t, const char *, const char *));
   const char *p;
   unsigned int c;
   unsigned int field_width = 0;
@@ -702,7 +688,7 @@ decode_one_format (const char *s_orig, const char *s, const char **next,
 	    size = sizeof (int);
 	  else
 	    {
-	      if (size > MAX_INTEGRAL_TYPE_SIZE
+	      if (MAX_INTEGRAL_TYPE_SIZE < size
 		  || integral_type_size[size] == NO_SIZE)
 		{
 		  error (0, 0, _("invalid type string `%s';\n\
@@ -986,15 +972,15 @@ skip (off_t n_skip)
       if (n_skip == 0)
 	break;
 
-      /* First try using fseek.  For large offsets, this extra work is
+      /* First try seeking.  For large offsets, this extra work is
 	 worthwhile.  If the offset is below some threshold it may be
 	 more efficient to move the pointer by reading.  There are two
-	 issues when trying to use fseek:
+	 issues when trying to seek:
 	   - the file must be seekable.
 	   - before seeking to the specified position, make sure
 	     that the new position is in the current file.
-	     Try to do that by getting file's size using fstat().
-	     But that will work only for regular files and dirs.  */
+	     Try to do that by getting file's size using fstat.
+	     But that will work only for regular files.  */
 
       if (fstat (fileno (in_stream), &file_stats))
 	{
@@ -1003,14 +989,14 @@ skip (off_t n_skip)
 	  continue;
 	}
 
-      /* The st_size field is valid only for regular files and
-	 directories.  FIXME: is the preceding true?
+      /* The st_size field is valid only for regular files
+	 (and for symbolic links, which cannot occur here).
 	 If the number of bytes left to skip is at least as large as
 	 the size of the current file, we can decrement
 	 n_skip and go on to the next file.  */
-      if (S_ISREG (file_stats.st_mode) || S_ISDIR (file_stats.st_mode))
+      if (S_ISREG (file_stats.st_mode))
 	{
-	  if (n_skip >= file_stats.st_size)
+	  if (file_stats.st_size <= n_skip)
 	    {
 	      n_skip -= file_stats.st_size;
 	      if (in_stream != stdin && fclose (in_stream) == EOF)
@@ -1022,10 +1008,7 @@ skip (off_t n_skip)
 	    }
 	  else
 	    {
-	      /* Try fseeko if available, fseek otherwise.  */
-	      if (fseeko (in_stream, n_skip, SEEK_SET) == 0
-		  || (n_skip <= LONG_MAX
-		      && fseek (in_stream, (long) n_skip, SEEK_SET) == 0))
+	      if (0 <= lseek (fileno (in_stream), n_skip, SEEK_CUR))
 		{
 		  n_skip = 0;
 		  break;
@@ -1036,10 +1019,10 @@ skip (off_t n_skip)
       /* Seek didn't work or wasn't attempted;  position the file pointer
 	 by reading.  */
 
-      for (j = n_skip / BUFSIZ; j >= 0; j--)
+      for (j = n_skip / BUFSIZ; 0 <= j; j--)
 	{
 	  char buf[BUFSIZ];
-	  size_t n_bytes_to_read = (j > 0
+	  size_t n_bytes_to_read = (0 < j
 				    ? BUFSIZ
 				    : n_skip % BUFSIZ);
 	  size_t n_bytes_read;
@@ -1060,13 +1043,13 @@ skip (off_t n_skip)
 }
 
 static const char *
-format_address_none (long unsigned int address)
+format_address_none (off_t address ATTRIBUTE_UNUSED)
 {
   return "";
 }
 
 static const char *
-format_address_std (long unsigned int address)
+format_address_std (off_t address)
 {
   const char *address_string;
 
@@ -1076,7 +1059,7 @@ format_address_std (long unsigned int address)
 }
 
 static const char *
-format_address_label (long unsigned int address)
+format_address_label (off_t address)
 {
   const char *address_string;
   assert (output_address_fmt_string != NULL);
@@ -1099,7 +1082,7 @@ format_address_label (long unsigned int address)
    only when it has not been padded to length BYTES_PER_BLOCK.  */
 
 static void
-write_block (long unsigned int current_offset, long unsigned int n_bytes,
+write_block (off_t current_offset, off_t n_bytes,
 	     const char *prev_block, const char *curr_block)
 {
   static int first = 1;
@@ -1264,7 +1247,7 @@ read_block (size_t n, char *block, size_t *n_bytes_in_buffer)
 {
   int err;
 
-  assert (n > 0 && n <= bytes_per_block);
+  assert (0 < n && n <= bytes_per_block);
 
   *n_bytes_in_buffer = 0;
 
@@ -1341,7 +1324,7 @@ parse_old_offset (const char *s)
   int radix;
   off_t offset;
   enum strtol_error s_err;
-  long unsigned int tmp;
+  uintmax_t tmp;
 
   if (*s == '\0')
     return -1;
@@ -1363,12 +1346,15 @@ parse_old_offset (const char *s)
 	radix = 8;
     }
 
-  s_err = xstrtoul (s, NULL, radix, &tmp, "Bb");
+  s_err = xstrtoumax (s, NULL, radix, &tmp, "Bb");
   if (s_err != LONGINT_OK)
     {
       STRTOL_FAIL_WARN (s, _("old-style offset"), s_err);
       return -1;
     }
+  if (OFF_T_MAX < tmp)
+    error (EXIT_FAILURE, 0,
+	   _("%s is larger than the maximum file size on this system"), s);
   offset = tmp;
   return offset;
 }
@@ -1485,15 +1471,14 @@ dump_strings (void)
   err = 0;
   while (1)
     {
-      unsigned int i;
+      size_t i;
       int c;
 
       /* See if the next `string_min' chars are all printing chars.  */
     tryline:
 
       if (limit_bytes_to_format
-	  && address >= (n_bytes_to_skip + max_bytes_to_format -
-			 (off_t) string_min))
+	  && address >= (n_bytes_to_skip + max_bytes_to_format - string_min))
 	break;
 
       for (i = 0; i < string_min; i++)
@@ -1590,6 +1575,26 @@ dump_strings (void)
   return err;
 }
 
+/* There must be exactly one %s format specifier in FORMAT_TEMPLATE.
+   Return the just-malloc'd result of using sprintf to insert
+   OFF_T_PRINTF_FORMAT_STRING into FORMAT_TEMPLATE.
+   Technically, the caller should free this memory, but IMHO it's not
+   worth it in this case.  */
+static char *
+expand_address_fmt (char const *format_template)
+{
+  size_t len = strlen (format_template);
+  char *fmt = xmalloc (len + 1);
+  /* Ensure that the literal we're inserting is no longer than the two-byte
+     string `%s' it's replacing.  There's also the %%, so technically we don't
+     even need the `+ 1' above.  */
+  assert (OFF_T_PRINTF_FORMAT_STRING[0] == 0
+	  || OFF_T_PRINTF_FORMAT_STRING[1] == 0
+	  || OFF_T_PRINTF_FORMAT_STRING[2] == 0);
+  sprintf (fmt, format_template, OFF_T_PRINTF_FORMAT_STRING);
+  return fmt;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1598,7 +1603,7 @@ main (int argc, char **argv)
   unsigned int i;
   unsigned int l_c_m;
   unsigned int address_pad_len;
-  unsigned long int desired_width;
+  unsigned long int desired_width IF_LINT (= 0);
   int width_specified = 0;
   int n_failed_decodes = 0;
   int err;
@@ -1641,7 +1646,7 @@ main (int argc, char **argv)
   n_specs_allocated = 5;
   spec = (struct tspec *) xmalloc (n_specs_allocated * sizeof (struct tspec));
 
-  output_address_fmt_string = "%07o";
+  output_address_fmt_string = expand_address_fmt ("%%07%so");
   format_address = format_address_std;
   address_pad_len = 7;
   flag_dump_strings = 0;
@@ -1649,7 +1654,7 @@ main (int argc, char **argv)
   while ((c = getopt_long (argc, argv, "abcdfhilos::xw::A:j:N:t:v",
 			   long_options, NULL)) != -1)
     {
-      unsigned long int tmp;
+      uintmax_t tmp;
       enum strtol_error s_err;
 
       switch (c)
@@ -1661,17 +1666,17 @@ main (int argc, char **argv)
 	  switch (optarg[0])
 	    {
 	    case 'd':
-	      output_address_fmt_string = "%07d";
+	      output_address_fmt_string = expand_address_fmt ("%%07%sd");
 	      format_address = format_address_std;
 	      address_pad_len = 7;
 	      break;
 	    case 'o':
-	      output_address_fmt_string = "%07o";
+	      output_address_fmt_string = expand_address_fmt ("%%07%so");
 	      format_address = format_address_std;
 	      address_pad_len = 7;
 	      break;
 	    case 'x':
-	      output_address_fmt_string = "%06x";
+	      output_address_fmt_string = expand_address_fmt ("%%06%sx");
 	      format_address = format_address_std;
 	      address_pad_len = 6;
 	      break;
@@ -1690,7 +1695,7 @@ it must be one character from [doxn]"),
 	  break;
 
 	case 'j':
-	  s_err = xstrtoul (optarg, NULL, 0, &tmp, "bkm");
+	  s_err = xstrtoumax (optarg, NULL, 0, &tmp, "bkm");
 	  n_bytes_to_skip = tmp;
 	  if (s_err != LONGINT_OK)
 	    STRTOL_FATAL_ERROR (optarg, _("skip argument"), s_err);
@@ -1699,17 +1704,15 @@ it must be one character from [doxn]"),
 	case 'N':
 	  limit_bytes_to_format = 1;
 
-	  /* FIXME: if off_t is long long and that's an 8-byte type,
-	     use xstrtouq here.  */
-	  s_err = xstrtoul (optarg, NULL, 0, &tmp, "bkm");
+	  s_err = xstrtoumax (optarg, NULL, 0, &tmp, "bkm");
 	  max_bytes_to_format = tmp;
 	  if (s_err != LONGINT_OK)
 	    STRTOL_FATAL_ERROR (optarg, _("limit argument"), s_err);
 
-	  if (tmp > LONG_MAX)
+	  if (OFF_T_MAX < tmp)
 	    error (EXIT_FAILURE, 0,
-		   _("specified number of bytes `%s' is larger than \
-the maximum\nrepresentable value of type `long'"), optarg);
+		   _("%s is larger than the maximum file size on this system"),
+		   optarg);
 	  break;
 
 	case 's':
@@ -1717,11 +1720,18 @@ the maximum\nrepresentable value of type `long'"), optarg);
 	    string_min = 3;
 	  else
 	    {
-	      s_err = xstrtoul (optarg, NULL, 0, &string_min, "bkm");
+	      s_err = xstrtoumax (optarg, NULL, 0, &tmp, "bkm");
 	      if (s_err != LONGINT_OK)
 		STRTOL_FATAL_ERROR (optarg, _("minimum string length"), s_err);
+
+	      /* The minimum string length may be no larger than SIZE_MAX,
+		 since we may allocate a buffer of this size.  */
+	      if (SIZE_MAX < tmp)
+		error (EXIT_FAILURE, 0, _("%s is too large"), optarg);
+
+	      string_min = tmp;
 	    }
-	  ++flag_dump_strings;
+	  flag_dump_strings = 1;
 	  break;
 
 	case 't':
@@ -1771,9 +1781,13 @@ the maximum\nrepresentable value of type `long'"), optarg);
 	    }
 	  else
 	    {
-	      s_err = xstrtoul (optarg, NULL, 10, &desired_width, "");
+	      uintmax_t w_tmp;
+	      s_err = xstrtoumax (optarg, NULL, 10, &w_tmp, "");
 	      if (s_err != LONGINT_OK)
 		STRTOL_FATAL_ERROR (optarg, _("width specification"), s_err);
+	      if (ULONG_MAX < w_tmp)
+		error (EXIT_FAILURE, 0, _("%s is too large"), optarg);
+	      desired_width = w_tmp;
 	    }
 	  break;
 
@@ -1875,7 +1889,7 @@ the maximum\nrepresentable value of type `long'"), optarg);
 
 	  if (output_address_fmt_string == NULL)
 	    {
-	      output_address_fmt_string = "(%07o)";
+	      output_address_fmt_string = expand_address_fmt ("(%%07%so)");
 	      format_address = format_address_std;
 	    }
 	  else
