@@ -3,12 +3,146 @@
    more time, I'll merge the remaining things in system.h and everything
    in this file will go back there. */
 
+#if STAT_MACROS_BROKEN
+# undef S_ISBLK
+# undef S_ISCHR
+# undef S_ISDIR
+# undef S_ISDOOR
+# undef S_ISFIFO
+# undef S_ISLNK
+# undef S_ISMPB
+# undef S_ISMPC
+# undef S_ISNWK
+# undef S_ISREG
+# undef S_ISSOCK
+#endif /* STAT_MACROS_BROKEN.  */
+
+#ifndef S_IFMT
+# define S_IFMT 0170000
+#endif
+#if !defined(S_ISBLK) && defined(S_IFBLK)
+# define S_ISBLK(m) (((m) & S_IFMT) == S_IFBLK)
+#endif
+#if !defined(S_ISCHR) && defined(S_IFCHR)
+# define S_ISCHR(m) (((m) & S_IFMT) == S_IFCHR)
+#endif
+#if !defined(S_ISDIR) && defined(S_IFDIR)
+# define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
+#endif
+#if !defined(S_ISREG) && defined(S_IFREG)
+# define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
+#endif
+#if !defined(S_ISFIFO) && defined(S_IFIFO)
+# define S_ISFIFO(m) (((m) & S_IFMT) == S_IFIFO)
+#endif
+#if !defined(S_ISLNK) && defined(S_IFLNK)
+# define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
+#endif
+#if !defined(S_ISSOCK) && defined(S_IFSOCK)
+# define S_ISSOCK(m) (((m) & S_IFMT) == S_IFSOCK)
+#endif
+#if !defined(S_ISMPB) && defined(S_IFMPB) /* V7 */
+# define S_ISMPB(m) (((m) & S_IFMT) == S_IFMPB)
+# define S_ISMPC(m) (((m) & S_IFMT) == S_IFMPC)
+#endif
+#if !defined(S_ISNWK) && defined(S_IFNWK) /* HP/UX */
+# define S_ISNWK(m) (((m) & S_IFMT) == S_IFNWK)
+#endif
+#if !defined(S_ISDOOR) && defined(S_IFDOOR) /* Solaris 2.5 and up */
+# define S_ISDOOR(m) (((m) & S_IFMT) == S_IFDOOR)
+#endif
+
+#if !S_ISUID
+# define S_ISUID 04000
+#endif
+#if !S_ISGID
+# define S_ISGID 02000
+#endif
+
+/* S_ISVTX is a common extension to POSIX.1.  */
+#ifndef S_ISVTX
+# define S_ISVTX 01000
+#endif
+
+#if !S_IRUSR && S_IREAD
+# define S_IRUSR S_IREAD
+#endif
+#if !S_IRUSR
+# define S_IRUSR 00400
+#endif
+#if !S_IRGRP
+# define S_IRGRP (S_IRUSR >> 3)
+#endif
+#if !S_IROTH
+# define S_IROTH (S_IRUSR >> 6)
+#endif
+
+#if !S_IWUSR && S_IWRITE
+# define S_IWUSR S_IWRITE
+#endif
+#if !S_IWUSR
+# define S_IWUSR 00200
+#endif
+#if !S_IWGRP
+# define S_IWGRP (S_IWUSR >> 3)
+#endif
+#if !S_IWOTH
+# define S_IWOTH (S_IWUSR >> 6)
+#endif
+
+#if !S_IXUSR && S_IEXEC
+# define S_IXUSR S_IEXEC
+#endif
+#if !S_IXUSR
+# define S_IXUSR 00100
+#endif
+#if !S_IXGRP
+# define S_IXGRP (S_IXUSR >> 3)
+#endif
+#if !S_IXOTH
+# define S_IXOTH (S_IXUSR >> 6)
+#endif
+
+#if !S_IRWXU
+# define S_IRWXU (S_IRUSR | S_IWUSR | S_IXUSR)
+#endif
+#if !S_IRWXG
+# define S_IRWXG (S_IRGRP | S_IWGRP | S_IXGRP)
+#endif
+#if !S_IRWXO
+# define S_IRWXO (S_IROTH | S_IWOTH | S_IXOTH)
+#endif
+
+/* S_IXUGO is a common extension to POSIX.1.  */
+#if !S_IXUGO
+# define S_IXUGO (S_IXUSR | S_IXGRP | S_IXOTH)
+#endif
+
+#ifndef S_IRWXUGO
+# define S_IRWXUGO (S_IRWXU | S_IRWXG | S_IRWXO)
+#endif
+
+/* All the mode bits that can be affected by chmod.  */
+#define CHMOD_MODE_BITS \
+  (S_ISUID | S_ISGID | S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO)
+
+#ifdef ST_MTIM_NSEC
+# define ST_TIME_CMP_NS(a, b, ns) ((a).ns < (b).ns ? -1 : (a).ns > (b).ns)
+#else
+# define ST_TIME_CMP_NS(a, b, ns) 0
+#endif
+#define ST_TIME_CMP(a, b, s, ns) \
+  ((a).s < (b).s ? -1 : (a).s > (b).s ? 1 : ST_TIME_CMP_NS(a, b, ns))
+#define ATIME_CMP(a, b) ST_TIME_CMP (a, b, st_atime, st_atim.ST_MTIM_NSEC)
+#define CTIME_CMP(a, b) ST_TIME_CMP (a, b, st_ctime, st_ctim.ST_MTIM_NSEC)
+#define MTIME_CMP(a, b) ST_TIME_CMP (a, b, st_mtime, st_mtim.ST_MTIM_NSEC)
+
 #ifndef RETSIGTYPE
 # define RETSIGTYPE void
 #endif
 
 #ifndef __GNUC__
-# ifdef HAVE_ALLOCA_H
+# if HAVE_ALLOCA_H
 #  include <alloca.h>
 # else
 #  ifdef _AIX
@@ -33,6 +167,10 @@ char *alloca ();
 # include <sys/exceptn.h>
 #endif
 
+#if HAVE_STDINT_H
+# include <stdint.h>
+#endif
+
 #include <ctype.h>
 
 /* Jim Meyering writes:
@@ -71,6 +209,9 @@ char *alloca ();
 # define ISGRAPH(c) (IN_CTYPE_DOMAIN (c) && isprint (c) && !isspace (c))
 #endif
 
+/* This is defined in <sys/euc.h> on at least Solaris2.6 systems.  */
+#undef ISPRINT
+
 #define ISPRINT(c) (IN_CTYPE_DOMAIN (c) && isprint (c))
 #define ISALNUM(c) (IN_CTYPE_DOMAIN (c) && isalnum (c))
 #define ISALPHA(c) (IN_CTYPE_DOMAIN (c) && isalpha (c))
@@ -131,40 +272,78 @@ char *alloca ();
 
 #define STREQ(a, b) (strcmp ((a), (b)) == 0)
 
-#ifndef HAVE_DECL_FREE
+#if !HAVE_DECL_FREE
 void free ();
 #endif
 
-#ifndef HAVE_DECL_MALLOC
+#if !HAVE_DECL_MALLOC
 char *malloc ();
 #endif
 
-#ifndef HAVE_DECL_MEMCHR
+#if !HAVE_DECL_MEMCHR
 char *memchr ();
 #endif
 
-#ifndef HAVE_DECL_REALLOC
+#if !HAVE_DECL_REALLOC
 char *realloc ();
 #endif
 
-#ifndef HAVE_DECL_STPCPY
+#if !HAVE_DECL_STPCPY
 # ifndef stpcpy
 char *stpcpy ();
 # endif
 #endif
 
-#ifndef HAVE_DECL_STRSTR
+#if !HAVE_DECL_STRNDUP
+char *strndup ();
+#endif
+
+#if !HAVE_DECL_STRSTR
 char *strstr ();
 #endif
 
-#ifndef HAVE_DECL_GETENV
+#if !HAVE_DECL_GETENV
 char *getenv ();
 #endif
 
-#ifndef HAVE_DECL_LSEEK
+#if !HAVE_DECL_LSEEK
 off_t lseek ();
 #endif
 
+/* This is needed on some AIX systems.  */
+#if !HAVE_DECL_STRTOUL
+unsigned long strtoul ();
+#endif
+
+/* This is needed on some AIX systems.  */
+#if !HAVE_DECL_STRTOULL && HAVE_UNSIGNED_LONG_LONG
+unsigned long long strtoull ();
+#endif
+
+#if !HAVE_DECL_GETLOGIN
+char *getlogin ();
+#endif
+
+#if !HAVE_DECL_TTYNAME
+char *ttyname ();
+#endif
+
+#if !HAVE_DECL_GETEUID
+uid_t geteuid ();
+#endif
+
+#if !HAVE_DECL_GETPWUID
+struct passwd *getpwuid ();
+#endif
+
+#if !HAVE_DECL_GETGRGID
+struct group *getgrgid ();
+#endif
+
+#if !HAVE_DECL_GETUID
+uid_t getuid ();
+#endif
+
 #include "xalloc.h"
 
 #if ! defined HAVE_MEMPCPY && ! defined mempcpy
@@ -177,57 +356,57 @@ off_t lseek ();
    are more efficient but not thread-safe.  That they're not thread-safe
    is fine since all these applications are single threaded.  */
 
-#ifdef HAVE_CLEARERR_UNLOCKED
+#if HAVE_CLEARERR_UNLOCKED
 # undef clearerr
 # define clearerr(S) clearerr_unlocked (S)
 #endif
 
-#ifdef HAVE_FEOF_UNLOCKED
+#if HAVE_FEOF_UNLOCKED
 # undef feof
 # define feof(S) feof_unlocked (S)
 #endif
 
-#ifdef HAVE_FERROR_UNLOCKED
+#if HAVE_FERROR_UNLOCKED
 # undef ferror
 # define ferror(S) ferror_unlocked (S)
 #endif
 
-#ifdef HAVE_FFLUSH_UNLOCKED
+#if HAVE_FFLUSH_UNLOCKED
 # undef fflush
 # define fflush(S) fflush_unlocked (S)
 #endif
 
-#ifdef HAVE_FPUTC_UNLOCKED
+#if HAVE_FPUTC_UNLOCKED
 # undef fputc
 # define fputc(C, S) fputc_unlocked (C, S)
 #endif
 
-#ifdef HAVE_FREAD_UNLOCKED
+#if HAVE_FREAD_UNLOCKED
 # undef fread
 # define fread(P, Z, N, S) fread_unlocked (P, Z, N, S)
 #endif
 
-#ifdef HAVE_FWRITE_UNLOCKED
+#if HAVE_FWRITE_UNLOCKED
 # undef fwrite
 # define fwrite(P, Z, N, S) fwrite_unlocked (P, Z, N, S)
 #endif
 
-#ifdef HAVE_GETC_UNLOCKED
+#if HAVE_GETC_UNLOCKED
 # undef getc
 # define getc(S) getc_unlocked (S)
 #endif
 
-#ifdef HAVE_GETCHAR_UNLOCKED
+#if HAVE_GETCHAR_UNLOCKED
 # undef getchar
 # define getchar(S) getchar_unlocked (S)
 #endif
 
-#ifdef HAVE_PUTC_UNLOCKED
+#if HAVE_PUTC_UNLOCKED
 # undef putc
 # define putc(C, S) putc_unlocked (C, S)
 #endif
 
-#ifdef HAVE_PUTCHAR_UNLOCKED
+#if HAVE_PUTCHAR_UNLOCKED
 # undef putchar
 # define putchar(C) putchar_unlocked (C)
 #endif
@@ -252,8 +431,14 @@ char *base_name PARAMS ((char const *));
 
 /* Factor out some of the common --help and --version processing code.  */
 
-#define GETOPT_HELP_CHAR 250
-#define GETOPT_VERSION_CHAR 251
+/* These enum values cannot possibly conflict with the option values
+   ordinarily used by commands, including CHAR_MAX + 1, etc.  Avoid
+   CHAR_MIN - 1, as it may equal -1, the getopt end-of-options value.  */
+enum
+{
+  GETOPT_HELP_CHAR = (CHAR_MIN - 2),
+  GETOPT_VERSION_CHAR = (CHAR_MIN - 3)
+};
 
 #define GETOPT_HELP_OPTION_DECL \
   "help", no_argument, 0, GETOPT_HELP_CHAR
@@ -271,7 +456,6 @@ char *base_name PARAMS ((char const *));
 #define case_GETOPT_VERSION_CHAR(Program_name, Authors)			\
   case GETOPT_VERSION_CHAR:						\
     version_etc (stdout, Program_name, GNU_PACKAGE, VERSION, Authors);	\
-    close_stdout ();							\
     exit (EXIT_SUCCESS);						\
     break;
 
@@ -339,6 +523,18 @@ char *base_name PARAMS ((char const *));
 # define ULONG_MAX TYPE_MAXIMUM (unsigned long)
 #endif
 
+#ifndef SIZE_MAX
+# define SIZE_MAX TYPE_MAXIMUM (size_t)
+#endif
+
+#ifndef OFF_T_MIN
+# define OFF_T_MIN TYPE_MINIMUM (off_t)
+#endif
+
+#ifndef OFF_T_MAX
+# define OFF_T_MAX TYPE_MAXIMUM (off_t)
+#endif
+
 #ifndef UID_T_MAX
 # define UID_T_MAX TYPE_MAXIMUM (uid_t)
 #endif
@@ -350,3 +546,43 @@ char *base_name PARAMS ((char const *));
 #ifndef PID_T_MAX
 # define PID_T_MAX TYPE_MAXIMUM (pid_t)
 #endif
+
+#ifndef CHAR_BIT
+# define CHAR_BIT 8
+#endif
+
+/* Use this to suppress gcc's `...may be used before initialized' warnings. */
+#ifdef lint
+# define IF_LINT(Code) Code
+#else
+# define IF_LINT(Code) /* empty */
+#endif
+
+#ifndef __attribute__
+# if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 8) || __STRICT_ANSI__
+#  define __attribute__(x)
+# endif
+#endif
+
+#ifndef ATTRIBUTE_NORETURN
+# define ATTRIBUTE_NORETURN __attribute__ ((__noreturn__))
+#endif
+
+#ifndef ATTRIBUTE_UNUSED
+# define ATTRIBUTE_UNUSED __attribute__ ((__unused__))
+#endif
+
+#if defined strdupa
+# define ASSIGN_STRDUPA(DEST, S)		\
+  do { DEST = strdupa(S); } while (0)
+#else
+# define ASSIGN_STRDUPA(DEST, S)		\
+  do						\
+    {						\
+      const char *s_ = (S);			\
+      size_t len_ = strlen (s_) + 1;		\
+      char *tmp_dest_ = (char *) alloca (len_);	\
+      DEST = memcpy (tmp_dest_, (s_), len_);	\
+    }						\
+  while (0)
+#endif
