@@ -1,5 +1,5 @@
 /* tail -- output the last part of file(s)
-   Copyright (C) 89, 90, 91, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 1989, 90, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@
 #include <signal.h>
 
 #include "system.h"
+#include "closeout.h"
 #include "argmatch.h"
 #include "error.h"
 #include "safe-read.h"
@@ -43,12 +44,10 @@
 #define AUTHORS \
   "Paul Rubin, David MacKenzie, Ian Lance Taylor, and Jim Meyering"
 
-#ifndef OFF_T_MIN
-# define OFF_T_MIN TYPE_MINIMUM (off_t)
-#endif
-
-#ifndef OFF_T_MAX
-# define OFF_T_MAX TYPE_MAXIMUM (off_t)
+#ifndef ENOSYS
+  /* Some systems don't have ENOSYS -- this should be a big enough
+     value that no valid errno value will match it.  */
+# define ENOSYS 99999
 #endif
 
 /* Number of items to tail.  */
@@ -78,6 +77,10 @@ enum Follow_mode
   Follow_descriptor = 2
 };
 
+/* The types of files for which tail works.  */
+#define IS_TAILABLE_FILE_TYPE(Mode) \
+  (S_ISREG (Mode) || S_ISFIFO (Mode) || S_ISCHR (Mode))
+
 static char const *const follow_mode_string[] =
 {
   "descriptor", "name", 0
@@ -103,14 +106,19 @@ struct File_spec
   dev_t dev;
   ino_t ino;
 
+  /* The specified name initially referred to a directory or some other
+     type for which tail isn't meaningful.  Unlike for a permission problem
+     (tailable, below) once this is set, the name is not checked ever again.  */
+  int ignore;
+
   /* See description of DEFAULT_MAX_N_... below.  */
   unsigned int n_unchanged_stats;
 
   /* See description of DEFAULT_MAX_N_... below.  */
   unsigned int n_consecutive_size_changes;
 
-  /* A file is tailable if it is a regular file or a fifo and it is
-     readable.  */
+  /* A file is tailable if it exists, is readable, and is of type
+     IS_TAILABLE_FILE_TYPE.  */
   int tailable;
 
   /* The value of errno seen last time we checked this file.  */
@@ -180,19 +188,35 @@ static pid_t pid;
 /* Nonzero if we have ever read standard input.  */
 static int have_read_stdin;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  RETRY_OPTION = CHAR_MAX + 1,
+  MAX_UNCHANGED_STATS_OPTION,
+
+  /* FIXME: remove this in 2001, unless someone can show a good
+     reason to keep it.  */
+  MAX_CONSECUTIVE_SIZE_CHANGES_OPTION,
+
+  PID_OPTION,
+  LONG_FOLLOW_OPTION
+};
+
 static struct option const long_options[] =
 {
   /* --allow-missing is deprecated; use --retry instead
      FIXME: remove it some day */
-  {"allow-missing", no_argument, NULL, CHAR_MAX + 1},
+  {"allow-missing", no_argument, NULL, RETRY_OPTION},
   {"bytes", required_argument, NULL, 'c'},
-  {"follow", optional_argument, NULL, 'f'},
+  {"follow", optional_argument, NULL, LONG_FOLLOW_OPTION},
   {"lines", required_argument, NULL, 'n'},
-  {"max-unchanged-stats", required_argument, NULL, CHAR_MAX + 2},
-  {"max-consecutive-size-changes", required_argument, NULL, CHAR_MAX + 3},
-  {"pid", required_argument, NULL, CHAR_MAX + 4},
+  {"max-unchanged-stats", required_argument, NULL, MAX_UNCHANGED_STATS_OPTION},
+  {"max-consecutive-size-changes", required_argument, NULL,
+   MAX_CONSECUTIVE_SIZE_CHANGES_OPTION},
+  {"pid", required_argument, NULL, PID_OPTION},
   {"quiet", no_argument, NULL, 'q'},
-  {"retry", no_argument, NULL, CHAR_MAX + 1},
+  {"retry", no_argument, NULL, RETRY_OPTION},
   {"silent", no_argument, NULL, 'q'},
   {"sleep-interval", required_argument, NULL, 's'},
   {"verbose", no_argument, NULL, 'v'},
@@ -226,13 +250,15 @@ With no FILE, or when FILE is -, read standard input.\n\
                              -f, --follow, and --follow=descriptor are\n\
                              equivalent\n\
   -n, --lines=N            output the last N lines, instead of the last %d\n\
-      --max-unchanged-stats=N see the texinfo documentation\n\
-                             (the default is %d)\n\
-      --max-consecutive-size-changes=N see the texinfo documentation\n\
-                             (the default is %d)\n\
+      --max-unchanged-stats=N\n\
+                           with --follow=name, reopen a FILE which has not\n\
+                             changed size after N (default %d) iterations\n\
+                             to see if it has been unlinked or renamed\n\
+                             (this is the usual case of rotated log files)\n\
       --pid=PID            with -f, terminate after process ID, PID dies\n\
   -q, --quiet, --silent    never output headers giving file names\n\
-  -s, --sleep-interval=S   with -f, sleep S seconds between iterations\n\
+  -s, --sleep-interval=S   with -f, each iteration lasts approximately S\n\
+                             (default 1) seconds\n\
   -v, --verbose            always output headers giving file names\n\
       --help               display this help and exit\n\
       --version            output version information and exit\n\
@@ -255,8 +281,7 @@ recreated by some other program.\n\
 \n\
 "),
 	      DEFAULT_N_LINES, DEFAULT_N_LINES,
-	      DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS,
-	      DEFAULT_MAX_N_CONSECUTIVE_SIZE_CHANGES
+	      DEFAULT_MAX_N_UNCHANGED_STATS_BETWEEN_OPENS
 	      );
       puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
     }
@@ -295,13 +320,11 @@ close_fd (int fd, const char *filename)
 }
 
 static void
-write_header (const char *pretty_filename, const char *comment)
+write_header (const char *pretty_filename)
 {
   static int first_file = 1;
 
-  printf ("%s==> %s%s%s <==\n", (first_file ? "" : "\n"), pretty_filename,
-	  (comment ? ": " : ""),
-	  (comment ? comment : ""));
+  printf ("%s==> %s <==\n", (first_file ? "" : "\n"), pretty_filename);
   first_file = 0;
 }
 
@@ -435,6 +458,7 @@ pipe_lines (const char *pretty_filename, int fd, long int n_lines)
   int i;			/* Index into buffers.  */
   int total_lines = 0;		/* Total number of newlines in all buffers.  */
   int errors = 0;
+  int nbytes;			/* Size of most recent read */
 
   first = last = (LBUFFER *) xmalloc (sizeof (LBUFFER));
   first->nbytes = first->nlines = 0;
@@ -442,7 +466,7 @@ pipe_lines (const char *pretty_filename, int fd, long int n_lines)
   tmp = (LBUFFER *) xmalloc (sizeof (LBUFFER));
 
   /* Input is always read into a fresh buffer.  */
-  while ((tmp->nbytes = safe_read (fd, tmp->buffer, BUFSIZ)) > 0)
+  while ((nbytes = tmp->nbytes = safe_read (fd, tmp->buffer, BUFSIZ)) > 0)
     {
       tmp->nlines = 0;
       tmp->next = NULL;
@@ -480,15 +504,19 @@ pipe_lines (const char *pretty_filename, int fd, long int n_lines)
 	    tmp = (LBUFFER *) xmalloc (sizeof (LBUFFER));
 	}
     }
-  if (tmp->nbytes == -1)
+
+  free ((char *) tmp);
+
+  if (nbytes < 0)
     {
       error (0, errno, "%s", pretty_filename);
       errors = 1;
-      free ((char *) tmp);
       goto free_lbuffers;
     }
 
-  free ((char *) tmp);
+  /* If the file is empty, then bail out.  */
+  if (last->nbytes == 0)
+    goto free_lbuffers;
 
   /* This prevents a core dump when the pipe contains no newlines.  */
   if (n_lines == 0)
@@ -509,7 +537,7 @@ pipe_lines (const char *pretty_filename, int fd, long int n_lines)
   /* Find the correct beginning, then print the rest of the file.  */
   if (total_lines > n_lines)
     {
-      char *cp;
+      const char *cp;
 
       /* Skip `total_lines' - `n_lines' newlines.  We made sure that
          `total_lines' - `n_lines' <= `tmp->nlines'.  */
@@ -730,15 +758,14 @@ recheck (struct File_spec *f)
 	  error (0, errno, "%s", pretty_name (f));
 	}
     }
-  else if (!S_ISREG (new_stats.st_mode)
-	   && !S_ISFIFO (new_stats.st_mode))
+  else if (!IS_TAILABLE_FILE_TYPE (new_stats.st_mode))
     {
       fail = 1;
       f->errnum = -1;
-      error (0, 0,
-	     _("`%s' has been replaced with a non-regular file;  \
-cannot follow end of non-regular file"),
+      error (0, 0, _("`%s' has been replaced with an untailable file;\
+ giving up on this name"),
 	     pretty_name (f));
+      f->ignore = 1;
     }
   else
     {
@@ -781,7 +808,17 @@ cannot follow end of non-regular file"),
     }
   else
     {
-      close_fd (fd, pretty_name (f));
+      if (f->fd == -1)
+	{
+	  /* This happens when one iteration finds the file missing,
+	     then the preceding <dev,inode> pair is reused as the
+	     file is recreated.  */
+	  new_file = 1;
+	}
+      else
+	{
+	  close_fd (fd, pretty_name (f));
+	}
     }
 
   if (new_file)
@@ -793,6 +830,7 @@ cannot follow end of non-regular file"),
       f->ino = new_stats.st_ino;
       f->n_unchanged_stats = 0;
       f->n_consecutive_size_changes = 0;
+      f->ignore = 0;
       /* FIXME: check lseek return value  */
       lseek (f->fd, f->size, SEEK_SET);
     }
@@ -839,6 +877,9 @@ tail_forever (struct File_spec *f, int nfiles)
 	{
 	  struct stat stats;
 
+	  if (f[i].ignore)
+	    continue;
+
 	  if (f[i].fd < 0)
 	    {
 	      recheck (&f[i]);
@@ -847,9 +888,9 @@ tail_forever (struct File_spec *f, int nfiles)
 
 	  if (fstat (f[i].fd, &stats) < 0)
 	    {
-	      error (0, errno, "%s", pretty_name (&f[i]));
 	      f[i].fd = -1;
 	      f[i].errnum = errno;
+	      error (0, errno, "%s", pretty_name (&f[i]));
 	      continue;
 	    }
 
@@ -889,7 +930,7 @@ tail_forever (struct File_spec *f, int nfiles)
 
 	  if (stats.st_size < f[i].size)
 	    {
-	      write_header (pretty_name (&f[i]), _("file truncated"));
+	      error (0, 0, _("%s: file truncated"), pretty_name (&f[i]));
 	      last = i;
 	      /* FIXME: check lseek return value  */
 	      lseek (f[i].fd, stats.st_size, SEEK_SET);
@@ -900,7 +941,7 @@ tail_forever (struct File_spec *f, int nfiles)
 	  if (i != last)
 	    {
 	      if (print_headers)
-		write_header (pretty_name (&f[i]), NULL);
+		write_header (pretty_name (&f[i]));
 	      last = i;
 	    }
 	  f[i].size += dump_remainder (pretty_name (&f[i]), f[i].fd,
@@ -1103,7 +1144,7 @@ tail_file (struct File_spec *f, off_t n_units)
   else
     {
       if (print_headers)
-	write_header (pretty_name (f), NULL);
+	write_header (pretty_name (f));
       errors = tail (pretty_name (f), fd, n_units);
       if (forever)
 	{
@@ -1111,17 +1152,20 @@ tail_file (struct File_spec *f, off_t n_units)
 	  /* FIXME: duplicate code */
 	  if (fstat (fd, &stats) < 0)
 	    {
-	      error (0, errno, "%s", pretty_name (f));
 	      errors = 1;
 	      f->errnum = errno;
+	      error (0, errno, "%s", pretty_name (f));
 	    }
-	  else if (!S_ISREG (stats.st_mode) && !S_ISFIFO (stats.st_mode))
+	  else if (!IS_TAILABLE_FILE_TYPE (stats.st_mode))
 	    {
-	      error (0, 0, _("%s: cannot follow end of non-regular file"),
+	      error (0, 0, _("%s: cannot follow end of this type of file;\
+ giving up on this name"),
 		     pretty_name (f));
 	      errors = 1;
 	      f->errnum = -1;
+	      f->ignore = 1;
 	    }
+
 	  if (errors)
 	    {
 	      close_fd (fd, pretty_name (f));
@@ -1135,6 +1179,7 @@ tail_file (struct File_spec *f, off_t n_units)
 	      f->ino = stats.st_ino;
 	      f->n_unchanged_stats = 0;
 	      f->n_consecutive_size_changes = 0;
+	      f->ignore = 0;
 	    }
 	}
       else
@@ -1311,7 +1356,7 @@ parse_options (int argc, char **argv,
   count_lines = 1;
   forever = from_start = print_headers = 0;
 
-  while ((c = getopt_long (argc, argv, "c:n:f::qs:v", long_options, NULL))
+  while ((c = getopt_long (argc, argv, "c:n:fqs:v", long_options, NULL))
 	 != -1)
     {
       switch (c)
@@ -1329,8 +1374,8 @@ parse_options (int argc, char **argv,
 
 	  {
 	    strtol_error s_err;
-	    unsigned long int tmp_ulong;
-	    s_err = xstrtoul (optarg, NULL, 10, &tmp_ulong, "bkm");
+	    uintmax_t n;
+	    s_err = xstrtoumax (optarg, NULL, 10, &n, "bkm");
 	    if (s_err == LONGINT_INVALID)
 	      {
 		error (EXIT_FAILURE, 0, "%s: %s", optarg,
@@ -1338,18 +1383,21 @@ parse_options (int argc, char **argv,
 			? _("invalid number of lines")
 			: _("invalid number of bytes")));
 	      }
-	    if (s_err != LONGINT_OK || tmp_ulong > OFF_T_MAX)
-	      {
-		error (EXIT_FAILURE, 0,
-		       _("%s: %s is so large that it is not representable"),
-		       optarg,
-		       c == 'n' ? _("number of lines") : _("number of bytes"));
-	      }
-	    *n_units = (off_t) tmp_ulong;
+
+	    if (s_err != LONGINT_OK)
+	      error (EXIT_FAILURE, 0,
+		     _("%s: is so large that it is not representable"), optarg);
+
+	    if (OFF_T_MAX < n)
+	      error (EXIT_FAILURE, 0,
+		     _("%s is larger than the maximum file size on this system"),
+		     optarg);
+	    *n_units = (off_t) n;
 	  }
 	  break;
 
 	case 'f':
+	case LONG_FOLLOW_OPTION:
 	  forever = 1;
 	  if (optarg == NULL)
 	    follow_mode = DEFAULT_FOLLOW_MODE;
@@ -1358,11 +1406,11 @@ parse_options (int argc, char **argv,
 				     follow_mode_string, follow_mode_map);
 	  break;
 
-	case CHAR_MAX + 1:
+	case RETRY_OPTION:
 	  reopen_inaccessible_files = 1;
 	  break;
 
-	case CHAR_MAX + 2:
+	case MAX_UNCHANGED_STATS_OPTION:
 	  /* --max-unchanged-stats=N */
 	  if (xstrtoul (optarg, NULL, 10,
 			&max_n_unchanged_stats_between_opens, "") != LONGINT_OK)
@@ -1373,7 +1421,7 @@ parse_options (int argc, char **argv,
 	    }
 	  break;
 
-	case CHAR_MAX + 3:
+	case MAX_CONSECUTIVE_SIZE_CHANGES_OPTION:
   	  /* --max-consecutive-size-changes=N */
 	  if (xstrtoul (optarg, NULL, 10,
 			&max_n_consecutive_size_changes_between_opens, "")
@@ -1385,7 +1433,7 @@ parse_options (int argc, char **argv,
 	    }
 	  break;
 
-	case CHAR_MAX + 4:
+	case PID_OPTION:
 	  {
 	    strtol_error s_err;
 	    unsigned long int tmp_ulong;
@@ -1435,6 +1483,11 @@ parse_options (int argc, char **argv,
   if (pid && !forever)
     error (0, 0,
 	   _("warning: PID ignored; --pid=PID is useful only when following"));
+  else if (pid && kill (pid, 0) != 0 && errno == ENOSYS)
+    {
+      error (0, 0, _("warning: --pid=PID is not supported on this system"));
+      pid = 0;
+    }
 }
 
 int
@@ -1456,6 +1509,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   have_read_stdin = 0;
 
   {
@@ -1508,13 +1563,15 @@ main (int argc, char **argv)
 
   if (forever)
     {
+      /* This fflush appears to be required only on Solaris2.7.  */
+      if (fflush (stdout) < 0)
+	error (EXIT_FAILURE, errno, _("write error"));
+
       SETVBUF (stdout, NULL, _IONBF, 0);
       tail_forever (F, n_files);
     }
 
   if (have_read_stdin && close (0) < 0)
     error (EXIT_FAILURE, errno, "-");
-  if (fclose (stdout) == EOF)
-    error (EXIT_FAILURE, errno, _("write error"));
   exit (exit_status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
